// The Boid class

class Boid {

  PVector loc;
  PVector vel;
  PVector acc;
  float r;
  float maxforce;    // Maximum steering force
  float maxspeed;    // Maximum speed

  Boid(float x, float y) {
    acc = new PVector(0,0);
    vel = new PVector(random(-1,1),random(-1,1));
    loc = new PVector(x,y);
    r = 2.0;
    maxspeed = 2.5;
    maxforce = 0.05;
  }
  
  Boid(float x, float y, float maxSpd, float maxF) {
    acc = new PVector(0,0);
    vel = new PVector(random(-1,1),random(-1,1));
    loc = new PVector(x,y);
    r = 2.0;
    maxspeed = maxSpd;
    maxforce = maxF;
  }
  
  Boid( PVector A, float maxSpd, float maxF) {
    acc = new PVector(0,0);
    vel = new PVector(random(-1,1),random(-1,1));
    loc = new PVector(A.x,A.y);
    r = 2.0;
    maxspeed = maxSpd;
    maxforce = maxF;
  }

  void run(ArrayList<Boid> boids) {
    flock(boids);
  }
  
  void update() {
    vel.add(acc);
    vel.limit(maxspeed);
    loc.add(vel);
    borders();
    if( renderBoid ) {
      render();
    }
  }

  void applyForce(PVector force) {
    // We could add mass here if we want A = F / M
    acc.add(force);
  }

  // We accumulate a new acceleration each time based on three rules
  void flock(ArrayList<Boid> boids) {
    PVector sep = separate(boids);   // Separation
    PVector ali = align(boids);      // Alignment
    PVector coh = cohesion(boids);   // Cohesion
    // Arbitrarily weight these forces
    sep.mult(1.90);  // 1.40 // 1.90
    ali.mult(0.75); // 1.02 // 0.75
    coh.mult(Kcoh); // Kcoh
    // Add the force vectors to acceleration
    acc.mult(0);
    applyForce(sep);
    applyForce(ali);
    applyForce(coh);
  }

  // A method that calculates and applies a steering force towards a target
  // STEER = DESIRED MINUS VELOCITY
  PVector seek(PVector target) {
    PVector desired = PVector.sub(target,loc);  // A vector pointing from the location to the target
    // Normalize desired and scale to maximum speed
    desired.normalize();
    desired.mult(maxspeed);
    // Steering = Desired minus Velocity
    PVector steer = PVector.sub(desired,vel);
    steer.limit(maxforce);  // Limit to maximum steering force
    return steer;
  }
  
  void render() {
    pushMatrix();
    if( renderMode == RENDER_WIREFRAME ) {
    // Draw a triangle rotated in the direction of velocity
      float theta = vel.heading2D() + radians(90);
      fill(200,100);
      stroke(255);
      translate(loc.x,loc.y);
      rotate(theta);
      beginShape(TRIANGLES);
      vertex(0, -r*2);
      vertex(-r, r*2);
      vertex(r, r*2);
      endShape();
    } else if( renderMode == RENDER_CLOUD ) {
      float theta = vel.heading2D();
      translate(loc.x,loc.y);
      rotate(theta);
      tint(200,200,255,32);
      //int x = round(bkgrnd.width*(constrain(loc.x,0,width-1)/float(width)));
      //int y = round(bkgrnd.height*(constrain(loc.y,0,height-1)/float(height)));
      //color c = bkgrnd.pixels[y*bkgrnd.width+x];
      //tint(c);
      image(boidImg,0,0,boidImg.width*2.5,boidImg.height*2.0);
    } else {
      float theta = vel.heading2D();
      translate(loc.x,loc.y);
      rotate(theta);
      stroke(232);
      fill(232);
      ellipse(0, 0, 1.5, 1);
    }
    popMatrix();
  }

  // Wraparound
  void borders() {
    if (loc.x < -r) loc.x = width+r;
    if (loc.y < -r) loc.y = height+r;
    if (loc.x > width+r) loc.x = -r;
    if (loc.y > height+r) loc.y = -r;
  }

  // Separation
  // Method checks for nearby boids and steers away
  PVector separateOLD (ArrayList<Boid> boids) {
    float desiredseparation = 25.0f;
    PVector steer = new PVector(0,0,0);
    int count = 0;
    // For every boid in the system, check if it's too close
    for (Boid other : boids) {
      float d = PVector.dist(loc,other.loc);
      // If the distance is greater than 0 and less than an arbitrary amount (0 when you are yourself)
      if ((d > 0) && (d < desiredseparation)) {
        // Calculate vector pointing away from neighbor
        PVector diff = PVector.sub(loc,other.loc);
        diff.normalize();
        diff.div(d);        // Weight by distance
        steer.add(diff);
        count++;            // Keep track of how many
      }
    }
    // Average -- divide by how many
    if (count > 0) {
      steer.div((float)count);
    }

    // As long as the vector is greater than 0
    if (steer.mag() > 0) {
      // Implement Reynolds: Steering = Desired - Velocity
      steer.normalize();
      steer.mult(maxspeed);
      steer.sub(vel);
      steer.limit(maxforce);
    }
    return steer;
  }
  PVector separate (ArrayList boids) {
    float desiredseparation = sepDist;
    PVector steer = new PVector(0,0,0);
    int count = 0;
    boolean predatorAround=false;
    // For every boid in the system, check if it's too close
    for (int i = 0 ; i < boids.size(); i++) {
      Boid other = (Boid) boids.get(i);
      float d = PVector.dist(loc,other.loc);
      // If the distance is greater than 0 and less than an arbitrary 
      // amount (0 when you are yourself)
      if ((d > 0) && (d < desiredseparation)) {
        // Calculate vector pointing away from neighbor
        PVector diff = PVector.sub(loc,other.loc);
        predatorAround = escapeVector(diff, other);
        diff.normalize();
        diff.div(d);        // Weight by distance
        steer.add(diff);
        count++;            // Keep track of how many
      }
      else if ((d > 0) && (d < predDist) && (other instanceof Predator)) {
        predatorAround = true;
      }
    }
    // Average -- divide by how many
    if (count > 0) {
      steer.div((float)count);
    }

    // As long as the vector is greater than 0
    if (steer.mag() > 0) {
      steer.normalize();
      steer.mult(maxspeed);
      steer.sub(vel);
      steer.limit(maxforce);
      escapeSpeedUp(predatorAround, steer);
    }
    return steer;
  }
  
  void escapeSpeedUp(boolean predatorAround, PVector steer) {
    if (predatorAround) {
      steer.mult(10);
    }
  }
  //The boid needs to flee the predator
  boolean escapeVector(PVector diff, Boid other) {
    if (other instanceof Predator) {
      int mu=15;
      //diff.mult(mu);
      return true;
    }
    return false;
  }
  
  // Alignment
  // For every nearby boid in the system, calculate the average velocity
  PVector align (ArrayList<Boid> boids) {
    float neighbordist = 50;
    PVector sum = new PVector(0,0);
    int count = 0;
    for (Boid other : boids) {
      if( !(other instanceof Predator) ) {
        float d = PVector.dist(loc,other.loc);
        if ((d > 0) && (d < neighbordist)) {
          sum.add(other.vel);
          count++;
        }
      }
    }
    if (count > 0) {
      sum.div((float)count);
      sum.normalize();
      //sum.mult(maxspeed);
      sum.mult(1.1);
      PVector steer = PVector.sub(sum,vel);
      steer.limit(maxforce);
      return steer;
    } else {
      return new PVector(0,0);
    }
  }

  // Cohesion
  // For the average location (i.e. center) of all nearby boids, calculate steering vector towards that location
  PVector cohesion (ArrayList<Boid> boids) {
    float neighbordist = 50;
    PVector sum = new PVector(0,0);   // Start with empty vector to accumulate all locations
    int count = 0;
    for (Boid other : boids) {
      if( !(other instanceof Predator) ) {
        float d = PVector.dist(loc,other.loc);
        if ((d > 0) && (d < neighbordist)) {
          sum.add(other.loc); // Add location
          count++;
        }
      } else {
        float d = PVector.dist(loc,other.loc);
        if ((d > 0) && (d < neighbordist)) {
          sum.sub(other.loc); // Add location
          count++;
        }
      }
    }
    if (count > 0) {
      sum.div(count);
      return seek(sum);  // Steer towards the location
    } else {
      return new PVector(0,0);
    }
  }
  
  PVector getLoc() { return loc; }
}


